"""
OpenAPI v3.1 schema types, created according to the specification:
https://github.com/OAI/OpenAPI-Specification/blob/master/versions/3.1.1.md

The type orders are according to the contents of the specification:
https://github.com/OAI/OpenAPI-Specification/blob/master/versions/3.1.1.md#table-of-contents
"""

from typing import TYPE_CHECKING

from openapi_pydantic.compat import PYDANTIC_V2

from .callback import Callback as Callback
from .components import Components as Components
from .contact import Contact as Contact
from .datatype import DataType as DataType
from .discriminator import Discriminator as Discriminator
from .encoding import Encoding as Encoding
from .example import Example as Example
from .external_documentation import ExternalDocumentation as ExternalDocumentation
from .header import Header as Header
from .info import Info as Info
from .license import License as License
from .link import Link as Link
from .media_type import MediaType as MediaType
from .oauth_flow import OAuthFlow as OAuthFlow
from .oauth_flows import OAuthFlows as OAuthFlows
from .open_api import OpenAPI as OpenAPI
from .operation import Operation as Operation
from .parameter import Parameter as Parameter
from .parameter import ParameterLocation as ParameterLocation
from .path_item import PathItem as PathItem
from .paths import Paths as Paths
from .reference import Reference as Reference
from .request_body import RequestBody as RequestBody
from .response import Response as Response
from .responses import Responses as Responses
from .schema import Schema as Schema
from .schema import schema_validate as schema_validate
from .security_requirement import SecurityRequirement as SecurityRequirement
from .security_scheme import SecurityScheme as SecurityScheme
from .server import Server as Server
from .server_variable import ServerVariable as ServerVariable
from .tag import Tag as Tag
from .xml import XML as XML

if TYPE_CHECKING:
    pass
elif PYDANTIC_V2:
    # resolve forward references
    Encoding.model_rebuild()
    OpenAPI.model_rebuild()
    Components.model_rebuild()
    Operation.model_rebuild()
else:
    # resolve forward references
    Encoding.update_forward_refs(Header=Header)
    Schema.update_forward_refs()
    Operation.update_forward_refs(PathItem=PathItem)



from typing import Callable

import pytest

from openapi_pydantic.compat import PYDANTIC_V2
from openapi_pydantic.v3 import v3_0, v3_1


@pytest.mark.parametrize("version", ["3.0.4", "3.1.1"])
def test_parse_with_callback(version: str) -> None:
    data = {
        "openapi": version,
        "info": {"title": "API with Callback", "version": ""},
        "paths": {
            "/create": {
                "post": {
                    "responses": {"200": {"description": "Success"}},
                    "callbacks": {
                        "event": {
                            "callback": {
                                "post": {
                                    "responses": {"200": {"description": "Success"}}
                                }
                            }
                        }
                    },
                }
            }
        },
    }

    if version == "3.0.4":
        model_validate_3_0: Callable[[dict], v3_0.OpenAPI] = getattr(
            v3_0.OpenAPI, "model_validate" if PYDANTIC_V2 else "parse_obj"
        )
        assert model_validate_3_0(data) == v3_0.OpenAPI(
            info=v3_0.Info(title="API with Callback", version=""),
            paths={
                "/create": v3_0.PathItem(
                    post=v3_0.Operation(
                        responses={"200": v3_0.Response(description="Success")},
                        callbacks={
                            "event": {
                                "callback": v3_0.PathItem(
                                    post=v3_0.Operation(
                                        responses={
                                            "200": v3_0.Response(description="Success")
                                        }
                                    )
                                )
                            }
                        },
                    )
                )
            },
        )
    else:
        model_validate_3_1: Callable[[dict], v3_1.OpenAPI] = getattr(
            v3_1.OpenAPI, "model_validate" if PYDANTIC_V2 else "parse_obj"
        )
        assert model_validate_3_1(data) == v3_1.OpenAPI(
            info=v3_1.Info(title="API with Callback", version=""),
            paths={
                "/create": v3_1.PathItem(
                    post=v3_1.Operation(
                        responses={"200": v3_1.Response(description="Success")},
                        callbacks={
                            "event": {
                                "callback": v3_1.PathItem(
                                    post=v3_1.Operation(
                                        responses={
                                            "200": v3_1.Response(description="Success")
                                        }
                                    )
                                )
                            }
                        },
                    )
                )
            },
        )



from typing import Any, Optional

from pydantic import BaseModel

from openapi_pydantic.compat import PYDANTIC_V2, ConfigDict, Extra

_examples = [
    {
        "summary": "A foo example",
        "value": {"foo": "bar"},
    },
    {
        "summary": "This is an example in XML",
        "externalValue": "http://example.org/examples/address-example.xml",
    },
    {
        "summary": "This is a text example",
        "externalValue": "http://foo.bar/examples/address-example.txt",
    },
]


class Example(BaseModel):
    summary: Optional[str] = None
    """
    Short description for the example.
    """

    description: Optional[str] = None
    """
    Long description for the example.
    [CommonMark syntax](https://spec.commonmark.org/) MAY be used for rich text 
    representation.
    """

    value: Optional[Any] = None
    """
    Embedded literal example.
    The `value` field and `externalValue` field are mutually exclusive.
    To represent examples of media types that cannot naturally represented in JSON or 
    YAML, use a string value to contain the example, escaping where necessary.
    """

    externalValue: Optional[str] = None
    """
    A URL that points to the literal example.
    This provides the capability to reference examples that cannot easily be included 
    in JSON or YAML documents.
    
    The `value` field and `externalValue` field are mutually exclusive.
    See the rules for resolving [Relative References](#relativeReferencesURI).
    """

    if PYDANTIC_V2:
        model_config = ConfigDict(
            extra="allow",
            json_schema_extra={"examples": _examples},
        )

    else:

        class Config:
            extra = Extra.allow
            schema_extra = {"examples": _examples}



import logging

from .v3 import XML as XML
from .v3 import Callback as Callback
from .v3 import Components as Components
from .v3 import Contact as Contact
from .v3 import DataType as DataType
from .v3 import Discriminator as Discriminator
from .v3 import Encoding as Encoding
from .v3 import Example as Example
from .v3 import ExternalDocumentation as ExternalDocumentation
from .v3 import Header as Header
from .v3 import Info as Info
from .v3 import License as License
from .v3 import Link as Link
from .v3 import MediaType as MediaType
from .v3 import OAuthFlow as OAuthFlow
from .v3 import OAuthFlows as OAuthFlows
from .v3 import OpenAPI as OpenAPI
from .v3 import Operation as Operation
from .v3 import Parameter as Parameter
from .v3 import ParameterLocation as ParameterLocation
from .v3 import PathItem as PathItem
from .v3 import Paths as Paths
from .v3 import Reference as Reference
from .v3 import RequestBody as RequestBody
from .v3 import Response as Response
from .v3 import Responses as Responses
from .v3 import Schema as Schema
from .v3 import SecurityRequirement as SecurityRequirement
from .v3 import SecurityScheme as SecurityScheme
from .v3 import Server as Server
from .v3 import ServerVariable as ServerVariable
from .v3 import Tag as Tag
from .v3 import parse_obj as parse_obj
from .v3 import schema_validate as schema_validate

logging.getLogger(__name__).addHandler(logging.NullHandler())



from typing import TYPE_CHECKING, Dict

if TYPE_CHECKING:
    from .path_item import PathItem


Callback = Dict[str, "PathItem"]
"""
A map of possible out-of band callbacks related to the parent operation.
Each value in the map is a [Path Item Object](#pathItemObject)
that describes a set of requests that may be initiated by the API provider and the 
expected responses. The key value used to identify the path item object is an 
expression, evaluated at runtime, that identifies a URL to use for the callback 
operation.
"""

"""Patterned Fields"""

# {expression}: 'PathItem' = ...
"""
A Path Item Object used to define a callback request and expected responses.

A [complete example](../examples/v3.0/callback-example.yaml) is available.
"""



from typing import Any, Dict, Optional, Union

from pydantic import BaseModel, Field

from openapi_pydantic.compat import PYDANTIC_V2, ConfigDict, Extra

from .encoding import Encoding
from .example import Example
from .reference import Reference
from .schema import Schema

_examples = [
    {
        "schema": {"$ref": "#/components/schemas/Pet"},
        "examples": {
            "cat": {
                "summary": "An example of a cat",
                "value": {
                    "name": "Fluffy",
                    "petType": "Cat",
                    "color": "White",
                    "gender": "male",
                    "breed": "Persian",
                },
            },
            "dog": {
                "summary": "An example of a dog with a cat's name",
                "value": {
                    "name": "Puma",
                    "petType": "Dog",
                    "color": "Black",
                    "gender": "Female",
                    "breed": "Mixed",
                },
            },
        },
    }
]


class MediaType(BaseModel):
    """Each Media Type Object provides schema and examples for the media type
    identified by its key."""

    media_type_schema: Optional[Union[Reference, Schema]] = Field(
        default=None, alias="schema"
    )
    """
    The schema defining the content of the request, response, or parameter.
    """

    example: Optional[Any] = None
    """
    Example of the media type.
    
    The example object SHOULD be in the correct format as specified by the media type.
    
    The `example` field is mutually exclusive of the `examples` field.
    
    Furthermore, if referencing a `schema` which contains an example,
    the `example` value SHALL _override_ the example provided by the schema.
    """

    examples: Optional[Dict[str, Union[Example, Reference]]] = None
    """
    Examples of the media type.
    
    Each example object SHOULD match the media type and specified schema if present.
    
    The `examples` field is mutually exclusive of the `example` field.
    
    Furthermore, if referencing a `schema` which contains an example,
    the `examples` value SHALL _override_ the example provided by the schema.
    """

    encoding: Optional[Dict[str, Encoding]] = None
    """
    A map between a property name and its encoding information.
    The key, being the property name, MUST exist in the schema as a property.
    The encoding object SHALL only apply to `requestBody` objects
    when the media type is `multipart` or `application/x-www-form-urlencoded`.
    """

    if PYDANTIC_V2:
        model_config = ConfigDict(
            extra="allow",
            populate_by_name=True,
            json_schema_extra={"examples": _examples},
        )

    else:

        class Config:
            extra = Extra.allow
            allow_population_by_field_name = True
            schema_extra = {"examples": _examples}



from typing import Optional

from pydantic import BaseModel

from openapi_pydantic.compat import PYDANTIC_V2, ConfigDict, Extra

_examples = [
    {
        "name": "Apache 2.0",
        "url": "https://www.apache.org/licenses/LICENSE-2.0.html",
    }
]


class License(BaseModel):
    """
    License information for the exposed API.
    """

    name: str
    """
    **REQUIRED**. The license name used for the API.
    """

    url: Optional[str] = None
    """
    A URL to the license used for the API.
    MUST be in the format of a URL.
    """

    if PYDANTIC_V2:
        model_config = ConfigDict(
            extra="allow",
            json_schema_extra={"examples": _examples},
        )

    else:

        class Config:
            extra = Extra.allow
            schema_extra = {"examples": _examples}



# mypy: ignore-errors

import sys
from typing import Any, Optional

import pytest
from openapi_spec_validator import validate
from pydantic import BaseModel

from openapi_pydantic import (
    DataType,
    Header,
    Info,
    MediaType,
    OpenAPI,
    Operation,
    PathItem,
    RequestBody,
    Response,
    Schema,
)
from openapi_pydantic.compat import PYDANTIC_V2
from openapi_pydantic.util import PydanticSchema, construct_open_api_with_schema_class

if sys.version_info < (3, 9):
    from typing_extensions import Annotated, Literal
else:
    from typing import Annotated, Literal


def test_basic_schema() -> None:
    class SampleModel(BaseModel):
        required: bool
        optional: Optional[bool] = None
        one_literal_choice: Literal["only_choice"]
        multiple_literal_choices: Literal["choice1", "choice2"]

    part_api = construct_sample_api(SampleModel)

    api = construct_open_api_with_schema_class(part_api)
    assert api.components is not None
    assert api.components.schemas is not None

    if PYDANTIC_V2:
        json_api: Any = api.model_dump(mode="json", by_alias=True, exclude_none=True)
    else:
        json_api: Any = api.dict(by_alias=True, exclude_none=True)
    validate(json_api)


@pytest.mark.skipif(
    not PYDANTIC_V2,
    reason="Field-level JSON examples not supported in Pydantic V1",
)
def test_field_json_schema_example() -> None:
    from pydantic import WithJsonSchema

    Thing = Annotated[str, WithJsonSchema({"examples": ["thing1"]})]

    class SampleModel(BaseModel):
        a: Thing

    part_api = construct_sample_api(SampleModel)

    api = construct_open_api_with_schema_class(part_api)
    assert api.components is not None
    assert api.components.schemas is not None

    json_api: Any = api.model_dump(mode="json", by_alias=True, exclude_none=True)
    props = json_api["components"]["schemas"]["SampleRequest"]["properties"]
    assert props["a"]["examples"] == ["thing1"]

    validate(json_api)


def construct_sample_api(SampleModel) -> OpenAPI:
    class SampleRequest(SampleModel):
        model_config = {"json_schema_mode": "validation"}

    class SampleResponse(SampleModel):
        model_config = {"json_schema_mode": "serialization"}

    return OpenAPI(
        info=Info(
            title="Sample API",
            version="v0.0.1",
        ),
        paths={
            "/callme": PathItem(
                post=Operation(
                    requestBody=RequestBody(
                        content={
                            "application/json": MediaType(
                                schema=PydanticSchema(schema_class=SampleRequest)
                            )
                        }
                    ),
                    responses={
                        "200": Response(
                            description="resp",
                            headers={
                                "WWW-Authenticate": Header(
                                    description="Indicate how to authenticate",
                                    schema=Schema(type=DataType.STRING),
                                )
                            },
                            content={
                                "application/json": MediaType(
                                    schema=PydanticSchema(schema_class=SampleResponse)
                                )
                            },
                        )
                    },
                )
            )
        },
    )



from openapi_pydantic import SecurityScheme
from openapi_pydantic.compat import PYDANTIC_V2


def test_oidc_parsing() -> None:
    security_scheme_1 = SecurityScheme(
        type="openIdConnect", openIdConnectUrl="https://example.com/openIdConnect"
    )
    assert isinstance(security_scheme_1.openIdConnectUrl, str)
    dump1 = getattr(security_scheme_1, "model_dump_json" if PYDANTIC_V2 else "json")
    if PYDANTIC_V2:
        assert dump1(by_alias=True, exclude_none=True) == (
            '{"type":"openIdConnect","openIdConnectUrl":"https://example.com/openIdConnect"}'
        )
    else:
        assert dump1(by_alias=True, exclude_none=True) == (
            '{"type": "openIdConnect", "openIdConnectUrl": "https://example.com/openIdConnect"}'
        )

    security_scheme_2 = SecurityScheme(
        type="openIdConnect", openIdConnectUrl="/openIdConnect"
    )
    assert isinstance(security_scheme_2.openIdConnectUrl, str)
    dump2 = getattr(security_scheme_2, "model_dump_json" if PYDANTIC_V2 else "json")
    if PYDANTIC_V2:
        assert dump2(by_alias=True, exclude_none=True) == (
            '{"type":"openIdConnect","openIdConnectUrl":"/openIdConnect"}'
        )
    else:
        assert dump2(by_alias=True, exclude_none=True) == (
            '{"type": "openIdConnect", "openIdConnectUrl": "/openIdConnect"}'
        )

    security_scheme_3 = SecurityScheme(
        type="openIdConnect", openIdConnectUrl="openIdConnect"
    )
    assert isinstance(security_scheme_3.openIdConnectUrl, str)
    dump3 = getattr(security_scheme_3, "model_dump_json" if PYDANTIC_V2 else "json")
    if PYDANTIC_V2:
        assert dump3(by_alias=True, exclude_none=True) == (
            '{"type":"openIdConnect","openIdConnectUrl":"openIdConnect"}'
        )
    else:
        assert dump3(by_alias=True, exclude_none=True) == (
            '{"type": "openIdConnect", "openIdConnectUrl": "openIdConnect"}'
        )



"""Compatibility layer to make this package usable with Pydantic 1 or 2"""

from typing import TYPE_CHECKING, Dict, List, Optional, Tuple

from pydantic.version import VERSION as PYDANTIC_VERSION

__all__ = [
    "PYDANTIC_V2",
    "ConfigDict",
    "JsonSchemaMode",
    "models_json_schema",
    "RootModel",
    "Extra",
    "v1_schema",
    "DEFS_KEY",
    "min_length_arg",
]

PYDANTIC_MAJOR_VERSION = int(PYDANTIC_VERSION.split(".", 1)[0])
PYDANTIC_MINOR_VERSION = int(PYDANTIC_VERSION.split(".")[1])
PYDANTIC_V2 = PYDANTIC_MAJOR_VERSION >= 2

if TYPE_CHECKING:
    # Provide stubs for either version of Pydantic

    from enum import Enum
    from typing import Any, Literal, Type, TypedDict

    from pydantic import BaseModel
    from pydantic import ConfigDict as PydanticConfigDict

    def ConfigDict(
        extra: Literal["allow", "ignore", "forbid"] = "allow",
        json_schema_extra: Optional[Dict[str, Any]] = None,
        populate_by_name: bool = True,
    ) -> PydanticConfigDict:
        """Stub for pydantic.ConfigDict in Pydantic 2"""
        ...

    class Extra(Enum):
        """Stub for pydantic.Extra in Pydantic 1"""

        allow = "allow"
        ignore = "ignore"
        forbid = "forbid"

    class RootModel(BaseModel):
        """Stub for pydantic.RootModel in Pydantic 2"""

    JsonSchemaMode = Literal["validation", "serialization"]

    def models_json_schema(
        models: List[Tuple[Type[BaseModel], JsonSchemaMode]],
        *,
        by_alias: bool = True,
        ref_template: str = "#/$defs/{model}",
        schema_generator: Optional[type] = None,
    ) -> Tuple[Dict, Dict[str, Any]]:
        """Stub for pydantic.json_schema.models_json_schema in Pydantic 2"""
        ...

    def v1_schema(
        models: List[Type[BaseModel]],
        *,
        by_alias: bool = True,
        ref_prefix: str = "#/$defs",
    ) -> Dict[str, Any]:
        """Stub for pydantic.schema.schema in Pydantic 1"""
        ...

    DEFS_KEY = "$defs"

    class MinLengthArg(TypedDict):
        pass

    def min_length_arg(min_length: int) -> MinLengthArg:
        """Generate a min_length or min_items parameter for Field(...)"""
        ...

elif PYDANTIC_V2:
    from typing import TypedDict

    from pydantic import ConfigDict, RootModel
    from pydantic.json_schema import JsonSchemaMode, models_json_schema

    # Pydantic 2 renders JSON schemas using the keyword "$defs"
    DEFS_KEY = "$defs"

    class MinLengthArg(TypedDict):
        min_length: int

    def min_length_arg(min_length: int) -> MinLengthArg:
        return {"min_length": min_length}

    # Create V1 stubs. These should not be used when PYDANTIC_V2 is true.
    Extra = None
    v1_schema = None


else:
    from typing import TypedDict

    from pydantic import Extra
    from pydantic.schema import schema as v1_schema

    # Pydantic 1 renders JSON schemas using the keyword "definitions"
    DEFS_KEY = "definitions"

    class MinLengthArg(TypedDict):
        min_items: int

    def min_length_arg(min_length: int) -> MinLengthArg:
        return {"min_items": min_length}

    # Create V2 stubs. These should not be used when PYDANTIC_V2 is false.
    ConfigDict = None
    models_json_schema = None
    JsonSchemaMode = None
    RootModel = None



import logging
from typing import Callable, Generic, TypeVar

import pytest
from pydantic import BaseModel, Field

from openapi_pydantic import (
    Info,
    MediaType,
    OpenAPI,
    Operation,
    PathItem,
    Reference,
    RequestBody,
    Response,
)
from openapi_pydantic.compat import PYDANTIC_V2
from openapi_pydantic.util import PydanticSchema, construct_open_api_with_schema_class


def test_construct_open_api_with_schema_class_1() -> None:
    open_api = construct_base_open_api_1()
    result_open_api_1 = construct_open_api_with_schema_class(open_api)
    result_open_api_2 = construct_open_api_with_schema_class(
        open_api, [PingRequest, PingResponse]
    )
    assert result_open_api_1.components == result_open_api_2.components
    assert result_open_api_1 == result_open_api_2

    dump_json = getattr(result_open_api_1, "model_dump_json" if PYDANTIC_V2 else "json")
    open_api_json = dump_json(by_alias=True, exclude_none=True, indent=2)
    logging.debug(open_api_json)


def test_construct_open_api_with_schema_class_2() -> None:
    open_api_1 = construct_base_open_api_1()
    open_api_2 = construct_base_open_api_2()
    result_open_api_1 = construct_open_api_with_schema_class(open_api_1)
    result_open_api_2 = construct_open_api_with_schema_class(
        open_api_2, [PingRequest, PingResponse]
    )
    assert result_open_api_1 == result_open_api_2


def test_construct_open_api_with_schema_class_3() -> None:
    open_api_3 = construct_base_open_api_3()

    result_with_alias_1 = construct_open_api_with_schema_class(open_api_3)
    assert result_with_alias_1.components is not None
    assert result_with_alias_1.components.schemas is not None
    schema_with_alias = result_with_alias_1.components.schemas["PongResponse"]
    assert schema_with_alias.properties is not None
    assert "pong_foo" in schema_with_alias.properties
    assert "pong_bar" in schema_with_alias.properties

    result_with_alias_2 = construct_open_api_with_schema_class(
        open_api_3, by_alias=True
    )
    assert result_with_alias_1 == result_with_alias_2

    result_without_alias = construct_open_api_with_schema_class(
        open_api_3, by_alias=False
    )
    assert result_without_alias.components is not None
    assert result_without_alias.components.schemas is not None
    schema_without_alias = result_without_alias.components.schemas["PongResponse"]
    assert schema_without_alias.properties is not None
    assert "resp_foo" in schema_without_alias.properties
    assert "resp_bar" in schema_without_alias.properties


@pytest.mark.skipif(PYDANTIC_V2, reason="generic type for Pydantic V1")
def test_construct_open_api_with_schema_class_4_generic_response_v1() -> None:
    DataT = TypeVar("DataT")
    from pydantic.v1.generics import GenericModel

    class GenericResponse(GenericModel, Generic[DataT]):
        msg: str = Field(description="message of the generic response")
        data: DataT = Field(description="data value of the generic response")

    open_api_4 = construct_base_open_api_4_generic_response(
        GenericResponse[PongResponse]
    )

    result = construct_open_api_with_schema_class(open_api_4)
    assert result.components is not None
    assert result.components.schemas is not None
    assert "GenericResponse_PongResponse_" in result.components.schemas


@pytest.mark.skipif(not PYDANTIC_V2, reason="generic type for Pydantic V2")
def test_construct_open_api_with_schema_class_4_generic_response_v2() -> None:
    DataT = TypeVar("DataT")

    class GenericResponse(BaseModel, Generic[DataT]):
        msg: str = Field(description="message of the generic response")
        data: DataT = Field(description="data value of the generic response")

    open_api_4 = construct_base_open_api_4_generic_response(
        GenericResponse[PongResponse]
    )

    result = construct_open_api_with_schema_class(open_api_4)
    assert result.components is not None
    assert result.components.schemas is not None
    assert "GenericResponse_PongResponse_" in result.components.schemas


def construct_base_open_api_1() -> OpenAPI:
    model_validate: Callable[[dict], OpenAPI] = getattr(
        OpenAPI, "model_validate" if PYDANTIC_V2 else "parse_obj"
    )
    return model_validate(
        {
            "info": {"title": "My own API", "version": "v0.0.1"},
            "paths": {
                "/ping": {
                    "post": {
                        "requestBody": {
                            "content": {
                                "application/json": {
                                    "schema": PydanticSchema(schema_class=PingRequest)
                                }
                            }
                        },
                        "responses": {
                            "200": {
                                "description": "pong",
                                "content": {
                                    "application/json": {
                                        "schema": PydanticSchema(
                                            schema_class=PingResponse
                                        )
                                    }
                                },
                            }
                        },
                    }
                }
            },
        }
    )


def construct_base_open_api_2() -> OpenAPI:
    return OpenAPI(
        info=Info(
            title="My own API",
            version="v0.0.1",
        ),
        paths={
            "/ping": PathItem(
                post=Operation(
                    requestBody=RequestBody(
                        content={
                            "application/json": MediaType(
                                media_type_schema=Reference(
                                    **{"$ref": "#/components/schemas/PingRequest"}
                                )
                            )
                        }
                    ),
                    responses={
                        "200": Response(
                            description="pong",
                            content={
                                "application/json": MediaType(
                                    media_type_schema=Reference(
                                        **{"$ref": "#/components/schemas/PingResponse"}
                                    )
                                )
                            },
                        )
                    },
                )
            )
        },
    )


def construct_base_open_api_3() -> OpenAPI:
    return OpenAPI(
        info=Info(
            title="My own API",
            version="v0.0.1",
        ),
        paths={
            "/ping": PathItem(
                post=Operation(
                    requestBody=RequestBody(
                        content={
                            "application/json": MediaType(
                                media_type_schema=PydanticSchema(
                                    schema_class=PingRequest
                                )
                            )
                        }
                    ),
                    responses={
                        "200": Response(
                            description="pong",
                            content={
                                "application/json": MediaType(
                                    media_type_schema=PydanticSchema(
                                        schema_class=PongResponse
                                    )
                                )
                            },
                        )
                    },
                )
            )
        },
    )


def construct_base_open_api_4_generic_response(response_schema: type) -> OpenAPI:
    return OpenAPI(
        info=Info(
            title="My own API",
            version="v0.0.1",
        ),
        paths={
            "/ping": PathItem(
                post=Operation(
                    requestBody=RequestBody(
                        content={
                            "application/json": MediaType(
                                media_type_schema=PydanticSchema(
                                    schema_class=PingRequest
                                )
                            )
                        }
                    ),
                    responses={
                        "200": Response(
                            description="pong",
                            content={
                                "application/json": MediaType(
                                    media_type_schema=PydanticSchema(
                                        schema_class=response_schema
                                    )
                                )
                            },
                        )
                    },
                )
            )
        },
    )


class PingRequest(BaseModel):
    """Ping Request"""

    req_foo: str = Field(description="foo value of the request")
    req_bar: str = Field(description="bar value of the request")


class PingResponse(BaseModel):
    """Ping response"""

    resp_foo: str = Field(description="foo value of the response")
    resp_bar: str = Field(description="bar value of the response")


class PongResponse(BaseModel):
    """Pong response"""

    resp_foo: str = Field(alias="pong_foo", description="foo value of the response")
    resp_bar: str = Field(alias="pong_bar", description="bar value of the response")



from .parser import parse_obj as parse_obj
from .v3_1 import XML as XML
from .v3_1 import Callback as Callback
from .v3_1 import Components as Components
from .v3_1 import Contact as Contact
from .v3_1 import DataType as DataType
from .v3_1 import Discriminator as Discriminator
from .v3_1 import Encoding as Encoding
from .v3_1 import Example as Example
from .v3_1 import ExternalDocumentation as ExternalDocumentation
from .v3_1 import Header as Header
from .v3_1 import Info as Info
from .v3_1 import License as License
from .v3_1 import Link as Link
from .v3_1 import MediaType as MediaType
from .v3_1 import OAuthFlow as OAuthFlow
from .v3_1 import OAuthFlows as OAuthFlows
from .v3_1 import OpenAPI as OpenAPI
from .v3_1 import Operation as Operation
from .v3_1 import Parameter as Parameter
from .v3_1 import ParameterLocation as ParameterLocation
from .v3_1 import PathItem as PathItem
from .v3_1 import Paths as Paths
from .v3_1 import Reference as Reference
from .v3_1 import RequestBody as RequestBody
from .v3_1 import Response as Response
from .v3_1 import Responses as Responses
from .v3_1 import Schema as Schema
from .v3_1 import SecurityRequirement as SecurityRequirement
from .v3_1 import SecurityScheme as SecurityScheme
from .v3_1 import Server as Server
from .v3_1 import ServerVariable as ServerVariable
from .v3_1 import Tag as Tag
from .v3_1 import schema_validate as schema_validate



from typing import Dict, List, Union

from pydantic import BaseModel, Field
from typing_extensions import Literal

from openapi_pydantic import (
    Discriminator,
    Info,
    MediaType,
    OpenAPI,
    Operation,
    PathItem,
    Reference,
    RequestBody,
    Response,
    Schema,
)
from openapi_pydantic.compat import (
    DEFS_KEY,
    PYDANTIC_MINOR_VERSION,
    PYDANTIC_V2,
    models_json_schema,
    v1_schema,
)
from openapi_pydantic.util import PydanticSchema, construct_open_api_with_schema_class


class DataAModel(BaseModel):
    kind: Literal["a"]


class DataBModel(BaseModel):
    kind: Literal["b"]


class RequestModel(BaseModel):
    data: Union[DataAModel, DataBModel] = Field(discriminator="kind")


def construct_base_open_api() -> OpenAPI:
    return OpenAPI(
        info=Info(
            title="My own API",
            version="v0.0.1",
        ),
        paths={
            "/ping": PathItem(
                post=Operation(
                    requestBody=RequestBody(
                        content={
                            "application/json": MediaType(
                                media_type_schema=PydanticSchema(
                                    schema_class=RequestModel
                                )
                            )
                        }
                    ),
                    responses={"200": Response(description="pong")},
                )
            )
        },
    )


def test_pydantic_discriminator_schema_generation() -> None:
    """https://github.com/kuimono/openapi-schema-pydantic/issues/8"""

    a_kind: Dict[str, Union[str, List[str]]] = {"title": "Kind", "type": "string"}
    b_kind: Dict[str, Union[str, List[str]]] = {"title": "Kind", "type": "string"}

    if PYDANTIC_V2:
        _key_map, json_schema = models_json_schema([(RequestModel, "validation")])
        # In Pydantic v2, string literal types are mapped to consts.
        a_kind["const"] = "a"
        b_kind["const"] = "b"
        if PYDANTIC_MINOR_VERSION < 10:
            # Prior to 2.10, string literal types are also mapped to enums with a single entry.
            a_kind["enum"] = ["a"]
            b_kind["enum"] = ["b"]
    else:
        json_schema = v1_schema([RequestModel])
        # In Pydantic v1, string literal types are mapped to enums with a single entry.
        a_kind["enum"] = ["a"]
        b_kind["enum"] = ["b"]

    assert json_schema == {
        DEFS_KEY: {
            "DataAModel": {
                "properties": {
                    "kind": a_kind,
                },
                "required": ["kind"],
                "title": "DataAModel",
                "type": "object",
            },
            "DataBModel": {
                "properties": {
                    "kind": b_kind,
                },
                "required": ["kind"],
                "title": "DataBModel",
                "type": "object",
            },
            "RequestModel": {
                "properties": {
                    "data": {
                        "oneOf": [
                            {"$ref": f"#/{DEFS_KEY}/DataAModel"},
                            {"$ref": f"#/{DEFS_KEY}/DataBModel"},
                        ],
                        "discriminator": {
                            "mapping": {
                                "a": f"#/{DEFS_KEY}/DataAModel",
                                "b": f"#/{DEFS_KEY}/DataBModel",
                            },
                            "propertyName": "kind",
                        },
                        "title": "Data",
                    }
                },
                "required": ["data"],
                "title": "RequestModel",
                "type": "object",
            },
        }
    }


def test_pydantic_discriminator_openapi_generation() -> None:
    """https://github.com/kuimono/openapi-schema-pydantic/issues/8"""

    open_api = construct_open_api_with_schema_class(construct_base_open_api())
    assert open_api.components is not None
    assert open_api.components.schemas is not None
    json_schema = open_api.components.schemas["RequestModel"]
    assert json_schema.properties == {
        "data": Schema(
            oneOf=[
                Reference(
                    **{
                        "$ref": "#/components/schemas/DataAModel",
                        "summary": None,
                        "description": None,
                    }
                ),
                Reference(
                    **{
                        "$ref": "#/components/schemas/DataBModel",
                        "summary": None,
                        "description": None,
                    }
                ),
            ],
            title="Data",
            discriminator=Discriminator(
                propertyName="kind",
                mapping={
                    "a": "#/components/schemas/DataAModel",
                    "b": "#/components/schemas/DataBModel",
                },
            ),
        )
    }



from openapi_pydantic.compat import PYDANTIC_V2, ConfigDict, Extra

from .parameter import ParameterBase

_examples = [
    {
        "description": "The number of allowed requests in the current period",
        "schema": {"type": "integer"},
    }
]


class Header(ParameterBase):
    """
    The Header Object follows the structure of the
    [Parameter Object](#parameterObject) with the following changes:

    1. `name` MUST NOT be specified, it is given in the corresponding
        `headers` map.
    2. `in` MUST NOT be specified, it is implicitly in `header`.
    3. All traits that are affected by the location MUST be applicable
        to a location of `header` (for example, [`style`](#parameterStyle)).
    """

    if PYDANTIC_V2:
        model_config = ConfigDict(
            extra="allow",
            populate_by_name=True,
            json_schema_extra={"examples": _examples},
        )

    else:

        class Config:
            extra = Extra.allow
            allow_population_by_field_name = True
            schema_extra = {"examples": _examples}



from typing import Optional

from pydantic import BaseModel

from openapi_pydantic.compat import PYDANTIC_V2, ConfigDict, Extra

from .contact import Contact
from .license import License

_examples = [
    {
        "title": "Sample Pet Store App",
        "description": "This is a sample server for a pet store.",
        "termsOfService": "http://example.com/terms/",
        "contact": {
            "name": "API Support",
            "url": "http://www.example.com/support",
            "email": "support@example.com",
        },
        "license": {
            "name": "Apache 2.0",
            "url": "https://www.apache.org/licenses/LICENSE-2.0.html",
        },
        "version": "1.0.1",
    }
]


class Info(BaseModel):
    """
    The object provides metadata about the API.
    The metadata MAY be used by the clients if needed,
    and MAY be presented in editing or documentation generation tools for convenience.
    """

    title: str
    """
    **REQUIRED**. The title of the API.
    """

    description: Optional[str] = None
    """
    A short description of the API.
    [CommonMark syntax](https://spec.commonmark.org/) MAY be used for rich text 
    representation.
    """

    termsOfService: Optional[str] = None
    """
    A URL to the Terms of Service for the API.
    MUST be in the format of a URL.
    """

    contact: Optional[Contact] = None
    """
    The contact information for the exposed API.
    """

    license: Optional[License] = None
    """
    The license information for the exposed API.
    """

    version: str
    """
    **REQUIRED**. The version of the OpenAPI document
    (which is distinct from the [OpenAPI Specification version](#oasVersion) or the API 
    implementation version).
    """

    if PYDANTIC_V2:
        model_config = ConfigDict(
            extra="allow",
            json_schema_extra={"examples": _examples},
        )

    else:

        class Config:
            extra = Extra.allow
            schema_extra = {"examples": _examples}



from typing import Optional

from pydantic import BaseModel

from openapi_pydantic.compat import PYDANTIC_V2, ConfigDict, Extra

_examples = [
    {
        "name": "API Support",
        "url": "http://www.example.com/support",
        "email": "support@example.com",
    }
]


class Contact(BaseModel):
    """
    Contact information for the exposed API.
    """

    name: Optional[str] = None
    """
    The identifying name of the contact person/organization.
    """

    url: Optional[str] = None
    """
    The URL pointing to the contact information.
    MUST be in the form of a URL.
    """

    email: Optional[str] = None
    """
    The email address of the contact person/organization.
    MUST be in the form of an email address.
    """

    if PYDANTIC_V2:
        model_config = ConfigDict(
            extra="allow",
            json_schema_extra={"examples": _examples},
        )

    else:

        class Config:
            extra = Extra.allow
            schema_extra = {"examples": _examples}



from typing import List, Optional

from pydantic import BaseModel

from openapi_pydantic.compat import PYDANTIC_V2, ConfigDict, Extra


class ServerVariable(BaseModel):
    """An object representing a Server Variable for server URL template substitution."""

    enum: Optional[List[str]] = None
    """
    An enumeration of string values to be used if the substitution options are from a 
    limited set. The array SHOULD NOT be empty.
    """

    default: str
    """
    **REQUIRED**. The default value to use for substitution,
    which SHALL be sent if an alternate value is _not_ supplied.
    Note this behavior is different than the [Schema Object's](#schemaObject) treatment 
    of default values, because in those cases parameter values are optional.
    If the [`enum`](#serverVariableEnum) is defined, the value SHOULD exist in the 
    enum's values.
    """

    description: Optional[str] = None
    """
    An optional description for the server variable.
    [CommonMark syntax](https://spec.commonmark.org/) MAY be used for rich text 
    representation.
    """

    if PYDANTIC_V2:
        model_config = ConfigDict(
            extra="allow",
        )

    else:

        class Config:
            extra = Extra.allow



import logging

from pydantic import BaseModel

from openapi_pydantic import Reference, schema_validate
from openapi_pydantic.compat import (
    DEFS_KEY,
    PYDANTIC_V2,
    ConfigDict,
    Extra,
    models_json_schema,
    v1_schema,
)


def test_schema() -> None:
    schema = schema_validate(
        {
            "title": "reference list",
            "description": "schema for list of reference type",
            "allOf": [{"$ref": "#/definitions/TestType"}],
        }
    )
    logging.debug(f"schema.allOf={schema.allOf}")
    assert schema.allOf
    assert isinstance(schema.allOf, list)
    assert isinstance(schema.allOf[0], Reference)
    assert schema.allOf[0].ref == "#/definitions/TestType"


def test_additional_properties_is_bool() -> None:
    class TestModel(BaseModel):
        test_field: str

        if PYDANTIC_V2:
            model_config = ConfigDict(
                extra="forbid",
            )

        else:

            class Config:
                extra = Extra.forbid

    if PYDANTIC_V2:
        _key_map, schema_definition = models_json_schema([(TestModel, "validation")])
    else:
        schema_definition = v1_schema([TestModel])

    assert schema_definition == {
        DEFS_KEY: {
            "TestModel": {
                "title": "TestModel",
                "type": "object",
                "properties": {"test_field": {"title": "Test Field", "type": "string"}},
                "required": ["test_field"],
                "additionalProperties": False,
            }
        }
    }

    # allow "additionalProperties" to have boolean value
    result = schema_validate(schema_definition[DEFS_KEY]["TestModel"])
    assert result.additionalProperties is False



from typing import Any

from openapi_pydantic.compat import PYDANTIC_V2
from openapi_pydantic.v3.v3_0 import (
    XML,
    Callback,
    Components,
    Contact,
    Discriminator,
    Encoding,
    Example,
    ExternalDocumentation,
    Header,
    Info,
    License,
    Link,
    MediaType,
    OAuthFlow,
    OAuthFlows,
    OpenAPI,
    Operation,
    Parameter,
    PathItem,
    Paths,
    Reference,
    RequestBody,
    Response,
    Responses,
    Schema,
    SecurityRequirement,
    SecurityScheme,
    Server,
    ServerVariable,
    Tag,
)


def test_config_example() -> None:
    all_types = [
        OpenAPI,
        Info,
        Contact,
        License,
        Server,
        ServerVariable,
        Components,
        Paths,
        PathItem,
        Operation,
        ExternalDocumentation,
        Parameter,
        RequestBody,
        MediaType,
        Encoding,
        Responses,
        Response,
        Callback,
        Example,
        Link,
        Header,
        Tag,
        Reference,
        Schema,
        Discriminator,
        XML,
        SecurityScheme,
        OAuthFlows,
        OAuthFlow,
        SecurityRequirement,
    ]
    for schema_type in all_types:
        _assert_config_examples(schema_type)


def _assert_config_examples(schema_type: Any) -> None:
    if PYDANTIC_V2:
        if not hasattr(schema_type, "model_config"):
            return
        extra = schema_type.model_config.get("json_schema_extra")
        if extra is not None:
            examples = extra["examples"]
            for example_dict in examples:
                obj = schema_type.model_validate(example_dict)
                assert obj.model_fields_set

    else:
        Config = getattr(schema_type, "Config", None)
        schema_extra = getattr(Config, "schema_extra", None)
        if schema_extra is not None:
            examples = schema_extra["examples"]
            for example_dict in examples:
                obj = schema_type(**example_dict)
                assert obj.__fields_set__



from typing import Optional

from pydantic import BaseModel, Field

from openapi_pydantic.compat import PYDANTIC_V2, ConfigDict, Extra

from .oauth_flows import OAuthFlows

_examples = [
    {"type": "http", "scheme": "basic"},
    {"type": "apiKey", "name": "api_key", "in": "header"},
    {"type": "http", "scheme": "bearer", "bearerFormat": "JWT"},
    {
        "type": "oauth2",
        "flows": {
            "implicit": {
                "authorizationUrl": "https://example.com/api/oauth/dialog",
                "scopes": {
                    "write:pets": "modify pets in your account",
                    "read:pets": "read your pets",
                },
            }
        },
    },
    {
        "type": "openIdConnect",
        "openIdConnectUrl": "https://example.com/openIdConnect",
    },
    {
        "type": "openIdConnect",
        "openIdConnectUrl": "openIdConnect",
    },  # issue #5: allow relative path
]


class SecurityScheme(BaseModel):
    """
    Defines a security scheme that can be used by the operations.

    Supported schemes are HTTP authentication,
    an API key (either as a header, a cookie parameter or as a query parameter),
    mutual TLS (use of a client certificate),
    OAuth2's common flows (implicit, password, client credentials and authorization
    code) as defined in [RFC6749](https://tools.ietf.org/html/rfc6749),
    and [OpenID Connect Discovery](https://tools.ietf.org/html/draft-ietf-oauth-discovery-06).

    Please note that as of 2020, the implicit flow is about to be deprecated by
    [OAuth 2.0 Security Best Current Practice](https://tools.ietf.org/html/draft-ietf-oauth-security-topics).
    Recommended for most use case is Authorization Code Grant flow with PKCE.
    """

    type: str
    """
    **REQUIRED**. The type of the security scheme.
    Valid values are `"apiKey"`, `"http"`, "mutualTLS", `"oauth2"`, `"openIdConnect"`.
    """

    description: Optional[str] = None
    """
    A description for security scheme.
    [CommonMark syntax](https://spec.commonmark.org/) MAY be used for rich text 
    representation.
    """

    name: Optional[str] = None
    """
    **REQUIRED** for `apiKey`. The name of the header, query or cookie parameter to be 
    used.
    """

    security_scheme_in: Optional[str] = Field(alias="in", default=None)
    """
    **REQUIRED** for `apiKey`. The location of the API key. Valid values are `"query"`, 
    `"header"` or `"cookie"`.
    """

    scheme: Optional[str] = None
    """
    **REQUIRED** for `http`. The name of the HTTP Authorization scheme to be used in the
    [Authorization header as defined in RFC7235](https://tools.ietf.org/html/rfc7235#section-5.1).
    
    The values used SHOULD be registered in the
    [IANA Authentication Scheme registry](https://www.iana.org/assignments/http-authschemes/http-authschemes.xhtml).
    """

    bearerFormat: Optional[str] = None
    """
    A hint to the client to identify how the bearer token is formatted.
    
    Bearer tokens are usually generated by an authorization server,
    so this information is primarily for documentation purposes.
    """

    flows: Optional[OAuthFlows] = None
    """
    **REQUIRED** for `oauth2`. An object containing configuration information for the 
    flow types supported.
    """

    openIdConnectUrl: Optional[str] = None
    """
    **REQUIRED** for `openIdConnect`. OpenId Connect URL to discover OAuth2 
    configuration values. This MUST be in the form of a URL. The OpenID Connect 
    standard requires the use of TLS.
    """

    if PYDANTIC_V2:
        model_config = ConfigDict(
            extra="allow",
            populate_by_name=True,
            json_schema_extra={"examples": _examples},
        )

    else:

        class Config:
            extra = Extra.allow
            allow_population_by_field_name = True
            schema_extra = {"examples": _examples}



import enum
from typing import Any, Dict, Optional, Union

from pydantic import BaseModel, Field

from openapi_pydantic.compat import PYDANTIC_V2, ConfigDict, Extra

from .example import Example
from .media_type import MediaType
from .reference import Reference
from .schema import Schema

_examples = [
    {
        "name": "token",
        "in": "header",
        "description": "token to be passed as a header",
        "required": True,
        "schema": {
            "type": "array",
            "items": {"type": "integer", "format": "int64"},
        },
        "style": "simple",
    },
    {
        "name": "username",
        "in": "path",
        "description": "username to fetch",
        "required": True,
        "schema": {"type": "string"},
    },
    {
        "name": "id",
        "in": "query",
        "description": "ID of the object to fetch",
        "required": False,
        "schema": {"type": "array", "items": {"type": "string"}},
        "style": "form",
        "explode": True,
    },
    {
        "in": "query",
        "name": "freeForm",
        "schema": {
            "type": "object",
            "additionalProperties": {"type": "integer"},
        },
        "style": "form",
    },
    {
        "in": "query",
        "name": "coordinates",
        "content": {
            "application/json": {
                "schema": {
                    "type": "object",
                    "required": ["lat", "long"],
                    "properties": {
                        "lat": {"type": "number"},
                        "long": {"type": "number"},
                    },
                }
            }
        },
    },
]


class ParameterLocation(str, enum.Enum):
    """The location of a given parameter."""

    QUERY = "query"
    HEADER = "header"
    PATH = "path"
    COOKIE = "cookie"


class ParameterBase(BaseModel):
    """
    Base class for Parameter and Header.

    (Header is like Parameter, but has no `name` or `in` fields.)
    """

    description: Optional[str] = None
    """
    A brief description of the parameter.
    This could contain examples of use.
    [CommonMark syntax](https://spec.commonmark.org/) MAY be used for rich text 
    representation.
    """

    required: bool = False
    """
    Determines whether this parameter is mandatory.
    If the [parameter location](#parameterIn) is `"path"`, this property is 
    **REQUIRED** and its value MUST be `true`.
    Otherwise, the property MAY be included and its default value is `false`.
    """

    deprecated: bool = False
    """
    Specifies that a parameter is deprecated and SHOULD be transitioned out of usage.
    Default value is `false`.
    """

    style: Optional[str] = None
    """
    Describes how the parameter value will be serialized depending on the type of the 
    parameter value. Default values (based on value of `in`):
    
    - for `query` - `form`;
    - for `path` - `simple`;
    - for `header` - `simple`;
    - for `cookie` - `form`.
    """

    explode: Optional[bool] = None
    """
    When this is true, parameter values of type `array` or `object` generate separate 
    parameters for each value of the array or key-value pair of the map.
    For other types of parameters this property has no effect.
    When [`style`](#parameterStyle) is `form`, the default value is `true`.
    For all other styles, the default value is `false`.
    """

    param_schema: Optional[Union[Schema, Reference]] = Field(
        default=None, alias="schema"
    )
    """
    The schema defining the type used for the parameter.
    """

    example: Optional[Any] = None
    """
    Example of the parameter's potential value.
    The example SHOULD match the specified schema and encoding properties if present.
    The `example` field is mutually exclusive of the `examples` field.
    Furthermore, if referencing a `schema` that contains an example, 
    the `example` value SHALL _override_ the example provided by the schema.
    To represent examples of media types that cannot naturally be represented in JSON 
    or YAML, a string value can contain the example with escaping where necessary.
    """

    examples: Optional[Dict[str, Union[Example, Reference]]] = None
    """
    Examples of the parameter's potential value.
    Each example SHOULD contain a value in the correct format as specified in the 
    parameter encoding. The `examples` field is mutually exclusive of the `example` 
    field.
    Furthermore, if referencing a `schema` that contains an example,
    the `examples` value SHALL _override_ the example provided by the schema.
    """

    """
    For more complex scenarios, the [`content`](#parameterContent) property 
    can define the media type and schema of the parameter.
    A parameter MUST contain either a `schema` property, or a `content` property, but 
    not both.
    When `example` or `examples` are provided in conjunction with the `schema` object,
    the example MUST follow the prescribed serialization strategy for the parameter.
    """

    content: Optional[Dict[str, MediaType]] = None
    """
    A map containing the representations for the parameter.
    The key is the media type and the value describes it.
    The map MUST only contain one entry.
    """

    if PYDANTIC_V2:
        model_config = ConfigDict(
            extra="allow",
            populate_by_name=True,
            json_schema_extra={"examples": _examples},
        )

    else:

        class Config:
            extra = Extra.allow
            allow_population_by_field_name = True
            schema_extra = {"examples": _examples}


class Parameter(ParameterBase):
    """
    Describes a single operation parameter.

    A unique parameter is defined by a combination of a [name](#parameterName) and
    [location](#parameterIn).
    """

    """Fixed Fields"""

    name: str
    """
    **REQUIRED**. The name of the parameter.
    Parameter names are *case sensitive*.

    - If [`in`](#parameterIn) is `"path"`, the `name` field MUST correspond to a
      template expression occurring within the [path](#pathsPath) field in the
      [Paths Object](#pathsObject).
      See [Path Templating](#pathTemplating) for further information.
    - If [`in`](#parameterIn) is `"header"` and the `name` field is `"Accept"`,
      `"Content-Type"` or `"Authorization"`, the parameter definition SHALL be ignored.
    - For all other cases, the `name` corresponds to the parameter name used by the
      [`in`](#parameterIn) property.
    """

    param_in: ParameterLocation = Field(alias="in")
    """
    **REQUIRED**. The location of the parameter. Possible values are `"query"`,
    `"header"`, `"path"` or `"cookie"`.
    """

    allowEmptyValue: bool = False
    """
    Sets the ability to pass empty-valued parameters.
    This is valid only for `query` parameters and allows sending a parameter with an 
    empty value. Default value is `false`.
    If [`style`](#parameterStyle) is used, and if behavior is `n/a` (cannot be 
    serialized), the value of `allowEmptyValue` SHALL be ignored.
    Use of this property is NOT RECOMMENDED, as it is likely to be removed in a later 
    revision.
    """

    allowReserved: bool = False
    """
    Determines whether the parameter value SHOULD allow reserved characters,
    as defined by [RFC3986](https://tools.ietf.org/html/rfc3986#section-2.2)
    `:/?#[]@!$&'()*+,;=` to be included without percent-encoding.
    This property only applies to parameters with an `in` value of `query`.
    The default value is `false`.
    """



from typing import Dict, Optional

from pydantic import BaseModel

from openapi_pydantic.compat import PYDANTIC_V2, ConfigDict, Extra

from .media_type import MediaType

_examples = [
    {
        "description": "user to add to the system",
        "content": {
            "application/json": {
                "schema": {"$ref": "#/components/schemas/User"},
                "examples": {
                    "user": {
                        "summary": "User Example",
                        "externalValue": "http://foo.bar/examples/user-example.json",
                    }
                },
            },
            "application/xml": {
                "schema": {"$ref": "#/components/schemas/User"},
                "examples": {
                    "user": {
                        "summary": "User example in XML",
                        "externalValue": "http://foo.bar/examples/user-example.xml",
                    }
                },
            },
            "text/plain": {
                "examples": {
                    "user": {
                        "summary": "User example in Plain text",
                        "externalValue": "http://foo.bar/examples/user-example.txt",
                    }
                }
            },
            "*/*": {
                "examples": {
                    "user": {
                        "summary": "User example in other format",
                        "externalValue": "http://foo.bar/examples/user-example.whatever",
                    }
                }
            },
        },
    },
    {
        "description": "user to add to the system",
        "content": {
            "text/plain": {"schema": {"type": "array", "items": {"type": "string"}}}
        },
    },
]


class RequestBody(BaseModel):
    """Describes a single request body."""

    description: Optional[str] = None
    """
    A brief description of the request body.
    This could contain examples of use.  
    
    [CommonMark syntax](https://spec.commonmark.org/) MAY be used for rich text 
    representation.
    """

    content: Dict[str, MediaType]
    """
    **REQUIRED**. The content of the request body.
    The key is a media type or [media type range](https://tools.ietf.org/html/rfc7231#appendix-D)
    and the value describes it.
    
    For requests that match multiple keys, only the most specific key is applicable. 
    e.g. text/plain overrides text/*
    """

    required: bool = False
    """
    Determines if the request body is required in the request. Defaults to `false`.
    """

    if PYDANTIC_V2:
        model_config = ConfigDict(
            extra="allow",
            json_schema_extra={"examples": _examples},
        )

    else:

        class Config:
            extra = Extra.allow
            schema_extra = {"examples": _examples}



from typing import List, Literal, Optional

from pydantic import BaseModel

from openapi_pydantic.compat import PYDANTIC_V2, ConfigDict, Extra

from .components import Components
from .external_documentation import ExternalDocumentation
from .info import Info
from .paths import Paths
from .security_requirement import SecurityRequirement
from .server import Server
from .tag import Tag


class OpenAPI(BaseModel):
    """This is the root document object of the OpenAPI document."""

    openapi: Literal["3.0.4", "3.0.3", "3.0.2", "3.0.1", "3.0.0"] = "3.0.4"
    """
    **REQUIRED**. This string MUST be the [semantic version number](https://semver.org/spec/v2.0.0.html)
    of the [OpenAPI Specification version](#versions) that the OpenAPI document uses. 
    The `openapi` field SHOULD be used by tooling specifications and clients to 
    interpret the OpenAPI document. This is *not* related to the API 
    [`info.version`](#infoVersion) string.
    """

    info: Info
    """
    **REQUIRED**. Provides metadata about the API. The metadata MAY be used by tooling 
    as required.
    """

    servers: List[Server] = [Server(url="/")]
    """
    An array of Server Objects, which provide connectivity information to a target 
    server. If the `servers` property is not provided, or is an empty array,
    the default value would be a [Server Object](#serverObject) with a 
    [url](#serverUrl) value of `/`.
    """

    paths: Paths
    """
    **REQUIRED**. The available paths and operations for the API.
    """

    components: Optional[Components] = None
    """
    An element to hold various schemas for the specification.
    """

    security: Optional[List[SecurityRequirement]] = None
    """
    A declaration of which security mechanisms can be used across the API. 
    The list of values includes alternative security requirement objects that can be 
    used. Only one of the security requirement objects need to be satisfied to 
    authorize a request. Individual operations can override this definition. 
    To make security optional, an empty security requirement (`{}`) can be included in 
    the array.
    """

    tags: Optional[List[Tag]] = None
    """
    A list of tags used by the specification with additional metadata.
    The order of the tags can be used to reflect on their order by the parsing tools.
    Not all tags that are used by the [Operation Object](#operationObject) must be 
    declared. The tags that are not declared MAY be organized randomly or based on the 
    tools' logic. Each tag name in the list MUST be unique.
    """

    externalDocs: Optional[ExternalDocumentation] = None
    """
    Additional external documentation.
    """

    if PYDANTIC_V2:
        model_config = ConfigDict(
            extra="allow",
        )

    else:

        class Config:
            extra = Extra.allow



from typing import Dict, Optional

from pydantic import BaseModel

from openapi_pydantic.compat import PYDANTIC_V2, ConfigDict, Extra

_examples = [
    {
        "authorizationUrl": "https://example.com/api/oauth/dialog",
        "scopes": {
            "write:pets": "modify pets in your account",
            "read:pets": "read your pets",
        },
    },
    {
        "authorizationUrl": "https://example.com/api/oauth/dialog",
        "tokenUrl": "https://example.com/api/oauth/token",
        "scopes": {
            "write:pets": "modify pets in your account",
            "read:pets": "read your pets",
        },
    },
    {
        "authorizationUrl": "/api/oauth/dialog",
        "tokenUrl": "/api/oauth/token",
        "refreshUrl": "/api/oauth/token",
        "scopes": {
            "write:pets": "modify pets in your account",
            "read:pets": "read your pets",
        },
    },
]


class OAuthFlow(BaseModel):
    """
    Configuration details for a supported OAuth Flow
    """

    authorizationUrl: Optional[str] = None
    """
    **REQUIRED** for `oauth2 ("implicit", "authorizationCode")`.
    The authorization URL to be used for this flow.
    This MUST be in the form of a URL.
    """

    tokenUrl: Optional[str] = None
    """
    **REQUIRED** for `oauth2 ("password", "clientCredentials", "authorizationCode")`.
    The token URL to be used for this flow.
    This MUST be in the form of a URL.
    """

    refreshUrl: Optional[str] = None
    """
    The URL to be used for obtaining refresh tokens. This MUST be in the form of a URL.
    """

    scopes: Dict[str, str]
    """
    **REQUIRED**. The available scopes for the OAuth2 security scheme.
    A map between the scope name and a short description for it.
    The map MAY be empty.
    """

    if PYDANTIC_V2:
        model_config = ConfigDict(
            extra="allow",
            json_schema_extra={"examples": _examples},
        )

    else:

        class Config:
            extra = Extra.allow
            schema_extra = {"examples": _examples}



from typing import Any, Dict, Optional

from pydantic import BaseModel

from openapi_pydantic.compat import PYDANTIC_V2, ConfigDict, Extra

from .server import Server

_examples = [
    {
        "operationId": "getUserAddressByUUID",
        "parameters": {"userUuid": "$response.body#/uuid"},
    },
    {
        "operationRef": "#/paths/~12.0~1repositories~1{username}/get",
        "parameters": {"username": "$response.body#/username"},
    },
]


class Link(BaseModel):
    """
    The `Link object` represents a possible design-time link for a response.
    The presence of a link does not guarantee the caller's ability to successfully
    invoke it, rather it provides a known relationship and traversal mechanism between
    responses and other operations.

    Unlike _dynamic_ links (i.e. links provided **in** the response payload),
    the OAS linking mechanism does not require link information in the runtime response.

    For computing links, and providing instructions to execute them,
    a [runtime expression](#runtimeExpression) is used for accessing values in an
    operation and using them as parameters while invoking the linked operation.
    """

    operationRef: Optional[str] = None
    """
    A relative or absolute URI reference to an OAS operation.
    This field is mutually exclusive of the `operationId` field,
    and MUST point to an [Operation Object](#operationObject).
    Relative `operationRef` values MAY be used to locate an existing 
    [Operation Object](#operationObject) in the OpenAPI definition. See the rules for 
    resolving [Relative References](#relativeReferencesURI).
    """

    operationId: Optional[str] = None
    """
    The name of an _existing_, resolvable OAS operation, as defined with a unique 
    `operationId`.
    
    This field is mutually exclusive of the `operationRef` field.
    """

    parameters: Optional[Dict[str, Any]] = None
    """
    A map representing parameters to pass to an operation
    as specified with `operationId` or identified via `operationRef`.
    The key is the parameter name to be used,
    whereas the value can be a constant or an expression to be evaluated and passed to 
    the linked operation.
    
    The parameter name can be qualified using the [parameter location](#parameterIn) 
    `[{in}.]{name}` for operations that use the same parameter name in different 
    locations (e.g. path.id).
    """

    requestBody: Optional[Any] = None
    """
    A literal value or [{expression}](#runtimeExpression) to use as a request body when 
    calling the target operation.
    """

    description: Optional[str] = None
    """
    A description of the link.
    [CommonMark syntax](https://spec.commonmark.org/) MAY be used for rich text 
    representation.
    """

    server: Optional[Server] = None
    """
    A server object to be used by the target operation.
    """

    if PYDANTIC_V2:
        model_config = ConfigDict(
            extra="allow",
            json_schema_extra={"examples": _examples},
        )

    else:

        class Config:
            extra = Extra.allow
            schema_extra = {"examples": _examples}



from typing import Callable

from openapi_pydantic import (
    MediaType,
    Parameter,
    PathItem,
    Reference,
    Schema,
    SecurityScheme,
)
from openapi_pydantic.compat import PYDANTIC_V2

validate_func_name = "model_validate" if PYDANTIC_V2 else "parse_obj"


def test_media_type_alias() -> None:
    media_type_1 = MediaType(media_type_schema=Schema())
    media_type_2 = MediaType(schema=Schema())
    model_validate: Callable[[dict], MediaType] = getattr(MediaType, validate_func_name)
    media_type_3 = model_validate({"media_type_schema": Schema()})
    media_type_4 = model_validate({"schema": Schema()})
    assert media_type_1 == media_type_2 == media_type_3 == media_type_4


def test_parameter_alias() -> None:
    parameter_1 = Parameter(  # type: ignore
        name="test",
        param_in="path",
        param_schema=Schema(),
    )
    parameter_2 = Parameter(  # type: ignore
        name="test",
        param_in="path",
        schema=Schema(),
    )
    model_validate: Callable[[dict], Parameter] = getattr(Parameter, validate_func_name)
    parameter_3 = model_validate(
        {"name": "test", "param_in": "path", "param_schema": Schema()}
    )
    parameter_4 = model_validate({"name": "test", "in": "path", "schema": Schema()})
    assert parameter_1 == parameter_2 == parameter_3 == parameter_4


def test_path_item_alias() -> None:
    path_item_1 = PathItem(ref="#/dummy")
    model_validate: Callable[[dict], PathItem] = getattr(PathItem, validate_func_name)
    path_item_2 = model_validate({"ref": "#/dummy"})
    path_item_3 = model_validate({"$ref": "#/dummy"})
    assert path_item_1 == path_item_2 == path_item_3


def test_reference_alias() -> None:
    reference_1 = Reference(ref="#/dummy")  # type: ignore
    reference_2 = Reference(**{"$ref": "#/dummy"})
    model_validate: Callable[[dict], Reference] = getattr(Reference, validate_func_name)
    reference_3 = model_validate({"ref": "#/dummy"})
    reference_4 = model_validate({"$ref": "#/dummy"})
    assert reference_1 == reference_2 == reference_3 == reference_4


def test_security_scheme() -> None:
    security_scheme_1 = SecurityScheme(type="apiKey", security_scheme_in="header")
    model_validate: Callable[[dict], SecurityScheme] = getattr(
        SecurityScheme, validate_func_name
    )
    security_scheme_2 = model_validate(
        {"type": "apiKey", "security_scheme_in": "header"}
    )
    security_scheme_3 = model_validate({"type": "apiKey", "in": "header"})
    assert security_scheme_1 == security_scheme_2 == security_scheme_3


def test_schema() -> None:
    schema_1 = Schema(schema_not=Schema(), schema_format="email")
    model_validate: Callable[[dict], Schema] = getattr(Schema, validate_func_name)
    schema_2 = model_validate({"schema_not": Schema(), "schema_format": "email"})
    schema_3 = model_validate({"not": Schema(), "format": "email"})
    assert schema_1 == schema_2 == schema_3



from typing import Dict, Optional

from pydantic import Field

from openapi_pydantic.compat import PYDANTIC_V2, ConfigDict
from openapi_pydantic.v3.v3_0 import OpenAPI, Operation, PathItem


def test_swagger_openapi_v3() -> None:
    with open("tests/data/swagger_openapi_v3.0.1.json") as f:
        if PYDANTIC_V2:
            validate = getattr(ExtendedOpenAPI, "model_validate_json")
        else:
            validate = getattr(ExtendedOpenAPI, "parse_raw")
        open_api = validate(f.read())
    assert open_api


class ExtendedOperation(Operation):
    """Override classes to use "x-codegen-request-body-name" in Operation"""

    xCodegenRequestBodyName: Optional[str] = Field(
        default=None, alias="x-codegen-request-body-name"
    )

    if PYDANTIC_V2:
        model_config = ConfigDict(populate_by_name=True)

    else:

        class Config:
            allow_population_by_field_name = True


class ExtendedPathItem(PathItem):
    get: Optional[ExtendedOperation] = None
    put: Optional[ExtendedOperation] = None
    post: Optional[ExtendedOperation] = None
    delete: Optional[ExtendedOperation] = None
    options: Optional[ExtendedOperation] = None
    head: Optional[ExtendedOperation] = None
    patch: Optional[ExtendedOperation] = None
    trace: Optional[ExtendedOperation] = None


class ExtendedOpenAPI(OpenAPI):
    paths: Dict[str, ExtendedPathItem]  # type: ignore[assignment]



import logging
from typing import Callable

from openapi_pydantic import Info, OpenAPI, Operation, PathItem, Response
from openapi_pydantic.compat import PYDANTIC_V2


def test_readme_example() -> None:
    open_api_1 = readme_example_1()
    assert open_api_1
    dump_json = getattr(open_api_1, "model_dump_json" if PYDANTIC_V2 else "json")
    open_api_json_1 = dump_json(by_alias=True, exclude_none=True, indent=2)
    logging.debug(open_api_json_1)
    assert open_api_json_1

    open_api_2 = readme_example_2()
    assert open_api_1 == open_api_2

    open_api_3 = readme_example_3()
    assert open_api_1 == open_api_3


def readme_example_1() -> OpenAPI:
    """Construct OpenAPI using data class"""
    return OpenAPI(
        info=Info(
            title="My own API",
            version="v0.0.1",
        ),
        paths={
            "/ping": PathItem(
                get=Operation(responses={"200": Response(description="pong")})
            )
        },
    )


def readme_example_2() -> OpenAPI:
    """Construct OpenAPI from raw data object"""
    openapi_validate: Callable[[dict], OpenAPI] = getattr(
        OpenAPI, "model_validate" if PYDANTIC_V2 else "parse_obj"
    )
    return openapi_validate(
        {
            "info": {"title": "My own API", "version": "v0.0.1"},
            "paths": {
                "/ping": {"get": {"responses": {"200": {"description": "pong"}}}}
            },
        }
    )


def readme_example_3() -> OpenAPI:
    """Construct OpenAPI from mixed object"""
    openapi_validate: Callable[[dict], OpenAPI] = getattr(
        OpenAPI, "model_validate" if PYDANTIC_V2 else "parse_obj"
    )
    return openapi_validate(
        {
            "info": {"title": "My own API", "version": "v0.0.1"},
            "paths": {
                "/ping": PathItem(
                    get={"responses": {"200": Response(description="pong")}}
                )
            },
        }
    )



from typing import Dict, Optional

from pydantic import BaseModel

from openapi_pydantic.compat import PYDANTIC_V2, ConfigDict, Extra

_examples = [
    {
        "propertyName": "petType",
        "mapping": {
            "dog": "#/components/schemas/Dog",
            "monster": "https://gigantic-server.com/schemas/Monster/schema.json",
        },
    }
]


class Discriminator(BaseModel):
    """
    When request bodies or response payloads may be one of a number of different
    schemas, a `discriminator` object can be used to aid in serialization,
    deserialization, and validation.

    The discriminator is a specific object in a schema which is used to inform the
    consumer of the specification of an alternative schema based on the value
    associated with it.

    When using the discriminator, _inline_ schemas will not be considered.
    """

    propertyName: str
    """
    **REQUIRED**. The name of the property in the payload that will hold the 
    discriminator value.
    """

    mapping: Optional[Dict[str, str]] = None
    """
    An object to hold mappings between payload values and schema names or references.
    """

    if PYDANTIC_V2:
        model_config = ConfigDict(
            extra="allow",
            json_schema_extra={"examples": _examples},
        )

    else:

        class Config:
            extra = Extra.allow
            schema_extra = {"examples": _examples}



from enum import Enum


class DataType(str, Enum):
    """Data type of an object."""

    NULL = "null"
    STRING = "string"
    NUMBER = "number"
    INTEGER = "integer"
    BOOLEAN = "boolean"
    ARRAY = "array"
    OBJECT = "object"



from typing import Optional

from pydantic import BaseModel

from openapi_pydantic.compat import PYDANTIC_V2, ConfigDict, Extra

_examples = [{"description": "Find more info here", "url": "https://example.com"}]


class ExternalDocumentation(BaseModel):
    """Allows referencing an external resource for extended documentation."""

    description: Optional[str] = None
    """
    A short description of the target documentation.
    [CommonMark syntax](https://spec.commonmark.org/) MAY be used for rich text 
    representation.
    """

    url: str
    """
    **REQUIRED**. The URL for the target documentation.
    Value MUST be in the form of a URL.
    """

    if PYDANTIC_V2:
        model_config = ConfigDict(
            extra="allow",
            json_schema_extra={"examples": _examples},
        )

    else:

        class Config:
            extra = Extra.allow
            schema_extra = {"examples": _examples}



from typing import Optional

from pydantic import BaseModel

from openapi_pydantic.compat import PYDANTIC_V2, ConfigDict, Extra

from .external_documentation import ExternalDocumentation

_examples = [{"name": "pet", "description": "Pets operations"}]


class Tag(BaseModel):
    """
    Adds metadata to a single tag that is used by the
    [Operation Object](#operationObject).
    It is not mandatory to have a Tag Object per tag defined in the Operation Object
    instances.
    """

    name: str
    """
    **REQUIRED**. The name of the tag.
    """

    description: Optional[str] = None
    """
    A short description for the tag.
    [CommonMark syntax](https://spec.commonmark.org/) MAY be used for rich text 
    representation.
    """

    externalDocs: Optional[ExternalDocumentation] = None
    """
    Additional external documentation for this tag.
    """

    if PYDANTIC_V2:
        model_config = ConfigDict(
            extra="allow",
            json_schema_extra={"examples": _examples},
        )

    else:

        class Config:
            extra = Extra.allow
            schema_extra = {"examples": _examples}



import pytest
from pydantic import ValidationError

from openapi_pydantic.v3.v3_1 import Schema


@pytest.mark.parametrize(
    "datatype",
    (
        "string",
        "number",
        "integer",
        "boolean",
        "array",
        "object",
        "null",
    ),
)
def test_good_types_parse_and_equate(datatype: str) -> None:
    assert Schema(type=datatype).type == datatype


def test_bad_types_raise_validation_errors() -> None:
    with pytest.raises(ValidationError):
        Schema(type="invalid")

    with pytest.raises(ValidationError):
        Schema(anyOf=[{"type": "invalid"}])

    with pytest.raises(ValidationError):
        Schema(
            properties={
                "a": Schema(type="invalid"),
            },
        )



from typing import Literal

import pytest

from openapi_pydantic import parse_obj
from openapi_pydantic.v3 import v3_0, v3_1


@pytest.mark.parametrize("version", ["3.0.4", "3.0.3", "3.0.2", "3.0.1", "3.0.0"])
def test_parse_obj_3_0(
    version: Literal["3.0.4", "3.0.3", "3.0.2", "3.0.1", "3.0.0"]
) -> None:
    result = parse_obj(
        {
            "openapi": version,
            "info": {"title": "foo", "version": "0.1.0"},
            "paths": {"/": {}},
        }
    )

    assert result == v3_0.OpenAPI(
        openapi=version,
        info=v3_0.Info(title="foo", version="0.1.0"),
        paths={"/": v3_0.PathItem()},
    )


@pytest.mark.parametrize("version", ["3.1.1", "3.1.0"])
def test_parse_obj_3_1(version: Literal["3.1.1", "3.1.0"]) -> None:
    result = parse_obj(
        {
            "openapi": version,
            "info": {"title": "foo", "version": "0.1.0"},
            "paths": {"/": {}},
        }
    )

    assert result == v3_1.OpenAPI(
        openapi=version,
        info=v3_1.Info(title="foo", version="0.1.0"),
        paths={"/": v3_1.PathItem()},
    )



from typing import Dict, Optional, Union

from pydantic import BaseModel

from openapi_pydantic.compat import PYDANTIC_V2, ConfigDict, Extra

from .header import Header
from .link import Link
from .media_type import MediaType
from .reference import Reference

_examples = [
    {
        "description": "A complex object array response",
        "content": {
            "application/json": {
                "schema": {
                    "type": "array",
                    "items": {"$ref": "#/components/schemas/VeryComplexType"},
                }
            }
        },
    },
    {
        "description": "A simple string response",
        "content": {"text/plain": {"schema": {"type": "string"}}},
    },
    {
        "description": "A simple string response",
        "content": {"text/plain": {"schema": {"type": "string", "example": "whoa!"}}},
        "headers": {
            "X-Rate-Limit-Limit": {
                "description": "The number of allowed requests in the "
                "current period",
                "schema": {"type": "integer"},
            },
            "X-Rate-Limit-Remaining": {
                "description": "The number of remaining requests in the "
                "current period",
                "schema": {"type": "integer"},
            },
            "X-Rate-Limit-Reset": {
                "description": "The number of seconds left in the current period",
                "schema": {"type": "integer"},
            },
        },
    },
    {"description": "object created"},
]


class Response(BaseModel):
    """
    Describes a single response from an API Operation, including design-time,
    static `links` to operations based on the response.
    """

    description: str
    """
    **REQUIRED**. A short description of the response.
    [CommonMark syntax](https://spec.commonmark.org/) MAY be used for rich text 
    representation.
    """

    headers: Optional[Dict[str, Union[Header, Reference]]] = None
    """
    Maps a header name to its definition.
    [RFC7230](https://tools.ietf.org/html/rfc7230#page-22) states header names are case 
    insensitive.
    If a response header is defined with the name `"Content-Type"`, it SHALL be ignored.
    """

    content: Optional[Dict[str, MediaType]] = None
    """
    A map containing descriptions of potential response payloads.
    The key is a media type or [media type range](https://tools.ietf.org/html/rfc7231#appendix-D)
    and the value describes it.  
    
    For responses that match multiple keys, only the most specific key is applicable. 
    e.g. text/plain overrides text/*
    """

    links: Optional[Dict[str, Union[Link, Reference]]] = None
    """
    A map of operations links that can be followed from the response.
    The key of the map is a short name for the link, following the naming constraints 
    of the names for [Component Objects](#componentsObject).
    """

    if PYDANTIC_V2:
        model_config = ConfigDict(
            extra="allow",
            json_schema_extra={"examples": _examples},
        )

    else:

        class Config:
            extra = Extra.allow
            schema_extra = {"examples": _examples}



import enum


class DataType(str, enum.Enum):
    """Data type of an object.

    Note: OpenAPI 3.0.x does not support null as a data type.
    """

    STRING = "string"
    NUMBER = "number"
    INTEGER = "integer"
    BOOLEAN = "boolean"
    ARRAY = "array"
    OBJECT = "object"



from typing import Optional

from pydantic import BaseModel

from openapi_pydantic.compat import PYDANTIC_V2, ConfigDict, Extra

_examples = [
    {"name": "animal"},
    {"attribute": True},
    {"wrapped": True},
    {"namespace": "http://example.com/schema/sample", "prefix": "sample"},
    {"name": "aliens", "wrapped": True},
]


class XML(BaseModel):
    """
    A metadata object that allows for more fine-tuned XML model definitions.

    When using arrays, XML element names are *not* inferred (for singular/plural forms)
    and the `name` property SHOULD be used to add that information.
    See examples for expected behavior.
    """

    name: Optional[str] = None
    """
    Replaces the name of the element/attribute used for the described schema property.
    When defined within `items`, it will affect the name of the individual XML elements 
    within the list.
    When defined alongside `type` being `array` (outside the `items`),
    it will affect the wrapping element and only if `wrapped` is `true`.
    If `wrapped` is `false`, it will be ignored.
    """

    namespace: Optional[str] = None
    """
    The URI of the namespace definition.
    Value MUST be in the form of an absolute URI.
    """

    prefix: Optional[str] = None
    """
    The prefix to be used for the [name](#xmlName).
    """

    attribute: bool = False
    """
    Declares whether the property definition translates to an attribute instead of an 
    element. Default value is `false`.
    """

    wrapped: bool = False
    """
    MAY be used only for an array definition.
    Signifies whether the array is wrapped 
    (for example, `<books><book/><book/></books>`) or unwrapped (`<book/><book/>`).
    Default value is `false`.
    The definition takes effect only when defined alongside `type` being `array` 
    (outside the `items`).
    """

    if PYDANTIC_V2:
        model_config = ConfigDict(
            extra="allow",
            json_schema_extra={"examples": _examples},
        )

    else:

        class Config:
            extra = Extra.allow
            schema_extra = {"examples": _examples}



from pydantic import BaseModel, Field

from openapi_pydantic.compat import PYDANTIC_V2, ConfigDict, Extra

_examples = [
    {"$ref": "#/components/schemas/Pet"},
    {"$ref": "Pet.json"},
    {"$ref": "definitions.json#/Pet"},
]


class Reference(BaseModel):
    """
    A simple object to allow referencing other components in the specification.

    The Reference Object is defined by [JSON Reference](https://tools.ietf.org/html/draft-pbryan-zyp-json-ref-03)
    and follows the same structure, behavior and rules.

    For this specification, reference resolution is accomplished as defined by the JSON
    Reference specification and not by the JSON Schema specification.
    """

    ref: str = Field(alias="$ref")
    """**REQUIRED**. The reference string."""

    if PYDANTIC_V2:
        model_config = ConfigDict(
            extra="allow",
            populate_by_name=True,
            json_schema_extra={"examples": _examples},
        )

    else:

        class Config:
            extra = Extra.allow
            allow_population_by_field_name = True
            schema_extra = {"examples": _examples}



# mypy: ignore-errors

from typing import Optional

import pytest

from openapi_pydantic.compat import PYDANTIC_V2
from openapi_pydantic.v3.v3_0 import (
    Info,
    MediaType,
    OpenAPI,
    Operation,
    PathItem,
    RequestBody,
    Response,
    Schema,
)
from openapi_pydantic.v3.v3_0.util import (
    PydanticSchema,
    construct_open_api_with_schema_class,
)


@pytest.mark.skipif(not PYDANTIC_V2, reason="computed fields require Pydantic V2")
def test_optional_and_computed_fields() -> None:
    api = construct_sample_api()

    result = construct_open_api_with_schema_class(api)
    assert result.components is not None
    assert result.components.schemas is not None

    req_schema = result.components.schemas["SampleRequest"]
    assert isinstance(req_schema, Schema)
    assert req_schema.properties is not None
    assert req_schema.required is not None

    resp_schema = result.components.schemas["SampleResponse"]
    assert isinstance(resp_schema, Schema)
    assert resp_schema.properties is not None
    assert resp_schema.required is not None

    # When validating:
    # - required fields are still required
    # - optional fields are still optional
    # - computed fields don't exist
    assert "req" in req_schema.properties
    assert "opt" in req_schema.properties
    assert "comp" not in req_schema.properties
    assert set(req_schema.required) == {"req"}

    # When serializing:
    # - required fields are still required
    # - optional fields are still optional
    #   (except when json_schema_serialization_defaults_required is enabled)
    # - computed fields are required
    assert "req" in resp_schema.properties
    assert "opt" in resp_schema.properties
    assert "comp" in resp_schema.properties
    assert set(resp_schema.required) == {"req", "comp"}


def construct_sample_api() -> OpenAPI:
    from typing import TYPE_CHECKING, Callable

    from pydantic import BaseModel

    if TYPE_CHECKING:

        def computed_field(x: Callable) -> Callable: ...

    else:
        from pydantic import computed_field

    class SampleModel(BaseModel):
        req: bool
        opt: Optional[bool] = None

        @computed_field  # type: ignore
        @property
        def comp(self) -> bool:
            return True

    class SampleRequest(SampleModel):
        model_config = {"json_schema_mode": "validation"}

    class SampleResponse(SampleModel):
        model_config = {"json_schema_mode": "serialization"}

    return OpenAPI(
        info=Info(
            title="Sample API",
            version="v0.0.1",
        ),
        paths={
            "/callme": PathItem(
                post=Operation(
                    requestBody=RequestBody(
                        content={
                            "application/json": MediaType(
                                schema=PydanticSchema(schema_class=SampleRequest)
                            )
                        }
                    ),
                    responses={
                        "200": Response(
                            description="resp",
                            content={
                                "application/json": MediaType(
                                    schema=PydanticSchema(schema_class=SampleResponse)
                                )
                            },
                        )
                    },
                )
            )
        },
    )



from typing import Optional

from pydantic import BaseModel, Field

from openapi_pydantic.compat import PYDANTIC_V2, ConfigDict, Extra

_examples = [
    {"$ref": "#/components/schemas/Pet"},
    {"$ref": "Pet.json"},
    {"$ref": "definitions.json#/Pet"},
]


class Reference(BaseModel):
    """
    A simple object to allow referencing other components in the OpenAPI document.

    The `$ref` string value contains a URI [RFC3986](https://tools.ietf.org/html/rfc3986),
    which identifies the location of the value being referenced.

    See the rules for resolving [Relative References](#relativeReferencesURI).
    """

    ref: str = Field(alias="$ref")
    """**REQUIRED**. The reference identifier. This MUST be in the form of a URI."""

    summary: Optional[str] = None
    """
    A short summary which by default SHOULD override that of the referenced component.
    If the referenced object-type does not allow a `summary` field, then this field has 
    no effect.
    """

    description: Optional[str] = None
    """
    A description which by default SHOULD override that of the referenced component.
    [CommonMark syntax](https://spec.commonmark.org/) MAY be used for rich text 
    representation.
    If the referenced object-type does not allow a `description` field, then this field 
    has no effect.
    """

    if PYDANTIC_V2:
        model_config = ConfigDict(
            extra="allow",
            populate_by_name=True,
            json_schema_extra={"examples": _examples},
        )

    else:

        class Config:
            extra = Extra.allow
            allow_population_by_field_name = True
            schema_extra = {"examples": _examples}



from typing import TYPE_CHECKING, Any, Union

from pydantic import BaseModel, Field

from openapi_pydantic.compat import PYDANTIC_V2

from .v3_0 import OpenAPI as OpenAPIv3_0
from .v3_1 import OpenAPI as OpenAPIv3_1

OpenAPIv3 = Union[OpenAPIv3_1, OpenAPIv3_0]

if TYPE_CHECKING:

    def parse_obj(data: Any) -> OpenAPIv3:
        """Parse a raw object into an OpenAPI model with version inference."""
        ...

elif PYDANTIC_V2:
    from pydantic import RootModel

    class _OpenAPI(RootModel):
        root: OpenAPIv3 = Field(discriminator="openapi")

    def parse_obj(data: Any) -> OpenAPIv3:
        return _OpenAPI.model_validate(data).root

else:

    class _OpenAPI(BaseModel):
        __root__: OpenAPIv3 = Field(discriminator="openapi")

    def parse_obj(data: Any) -> OpenAPIv3:
        return _OpenAPI.parse_obj(data).__root__



import logging
import re
from typing import Any, Dict, Generic, List, Optional, Set, Type, TypeVar, cast

from pydantic import BaseModel

from openapi_pydantic.compat import (
    DEFS_KEY,
    PYDANTIC_V2,
    JsonSchemaMode,
    models_json_schema,
    v1_schema,
)

from . import Components, OpenAPI, Reference, Schema, schema_validate

logger = logging.getLogger(__name__)

PydanticType = TypeVar("PydanticType", bound=BaseModel)
ref_prefix = "#/components/schemas/"
ref_template = "#/components/schemas/{model}"


class PydanticSchema(Schema, Generic[PydanticType]):
    """Special `Schema` class to indicate a reference from pydantic class"""

    schema_class: Type[PydanticType]
    """the class that is used for generate the schema"""


def get_mode(
    cls: Type[BaseModel], default: JsonSchemaMode = "validation"
) -> JsonSchemaMode:
    """Get the JSON schema mode for a model class.

    The mode can be either "validation" or "serialization". In validation mode,
    computed fields are dropped and optional fields remain optional. In
    serialization mode, computed and optional fields are required.
    """
    if not hasattr(cls, "model_config"):
        return default
    mode = cls.model_config.get("json_schema_mode", default)
    if mode not in ("validation", "serialization"):
        raise ValueError(f"invalid json_schema_mode: {mode}")
    return cast(JsonSchemaMode, mode)


def construct_open_api_with_schema_class(
    open_api: OpenAPI,
    schema_classes: Optional[List[Type[BaseModel]]] = None,
    scan_for_pydantic_schema_reference: bool = True,
    by_alias: bool = True,
) -> OpenAPI:
    """
    Construct a new OpenAPI object, utilising pydantic classes to produce JSON schemas.

    :param open_api: the base `OpenAPI` object
    :param schema_classes: Pydantic classes that their schema will be used
                           "#/components/schemas" values
    :param scan_for_pydantic_schema_reference: flag to indicate if scanning for
                                               `PydanticSchemaReference` class
                                               is needed for "#/components/schemas"
                                               value updates
    :param by_alias: construct schema by alias (default is True)
    :return: new OpenAPI object with "#/components/schemas" values updated.
             If there is no update in "#/components/schemas" values, the original
             `open_api` will be returned.
    """
    copy_func = getattr(open_api, "model_copy" if PYDANTIC_V2 else "copy")
    new_open_api: OpenAPI = copy_func(deep=True)

    if scan_for_pydantic_schema_reference:
        extracted_schema_classes = _handle_pydantic_schema(new_open_api)
        if schema_classes:
            schema_classes = list({*schema_classes, *extracted_schema_classes})
        else:
            schema_classes = extracted_schema_classes

    if not schema_classes:
        return open_api

    schema_classes.sort(key=lambda x: x.__name__)
    logger.debug("schema_classes: %s", schema_classes)

    # update new_open_api with new #/components/schemas
    if PYDANTIC_V2:
        _key_map, schema_definitions = models_json_schema(
            [(c, get_mode(c)) for c in schema_classes],
            by_alias=by_alias,
            ref_template=ref_template,
        )
    else:
        schema_definitions = v1_schema(
            schema_classes, by_alias=by_alias, ref_prefix=ref_prefix
        )

    if not new_open_api.components:
        new_open_api.components = Components()
    if new_open_api.components.schemas:
        for existing_key in new_open_api.components.schemas:
            if existing_key in schema_definitions[DEFS_KEY]:
                logger.warning(
                    f'"{existing_key}" already exists in {ref_prefix}. '
                    f'The value of "{ref_prefix}{existing_key}" will be overwritten.'
                )
        new_open_api.components.schemas.update(_validate_schemas(schema_definitions))
    else:
        new_open_api.components.schemas = _validate_schemas(schema_definitions)
    return new_open_api


def _validate_schemas(schema_definitions: Dict[str, Any]) -> Dict[str, Schema]:
    """Convert JSON Schema definitions to parsed OpenAPI objects"""
    # Note: if an error occurs in schema_validate(), it may indicate that
    # the generated JSON schemas are not compatible with the version
    # of OpenAPI this module depends on.
    return {
        key: schema_validate(schema_dict)
        for key, schema_dict in schema_definitions[DEFS_KEY].items()
    }


def _handle_pydantic_schema(open_api: OpenAPI) -> List[Type[BaseModel]]:
    """
    This function traverses the `OpenAPI` object and

    1. Replaces the `PydanticSchema` object with `Reference` object, with correct ref
       value;
    2. Extracts the involved schema class from `PydanticSchema` object.

    **This function will mutate the input `OpenAPI` object.**

    :param open_api: the `OpenAPI` object to be traversed and mutated
    :return: a list of schema classes extracted from `PydanticSchema` objects
    """

    pydantic_types: Set[Type[BaseModel]] = set()

    def _traverse(obj: Any) -> None:
        if isinstance(obj, BaseModel):
            fields = getattr(
                obj, "model_fields_set" if PYDANTIC_V2 else "__fields_set__"
            )
            for field in fields:
                child_obj = obj.__getattribute__(field)
                if isinstance(child_obj, PydanticSchema):
                    logger.debug("PydanticSchema found in %s: %s", obj, child_obj)
                    obj.__setattr__(field, _construct_ref_obj(child_obj))
                    pydantic_types.add(child_obj.schema_class)
                else:
                    _traverse(child_obj)
        elif isinstance(obj, list):
            for index, elem in enumerate(obj):
                if isinstance(elem, PydanticSchema):
                    logger.debug(f"PydanticSchema found in list: {elem}")
                    obj[index] = _construct_ref_obj(elem)
                    pydantic_types.add(elem.schema_class)
                else:
                    _traverse(elem)
        elif isinstance(obj, dict):
            for key, value in obj.items():
                if isinstance(value, PydanticSchema):
                    logger.debug(f"PydanticSchema found in dict: {value}")
                    obj[key] = _construct_ref_obj(value)
                    pydantic_types.add(value.schema_class)
                else:
                    _traverse(value)

    _traverse(open_api)
    return list(pydantic_types)


def _construct_ref_obj(pydantic_schema: PydanticSchema[PydanticType]) -> Reference:
    """
    Construct a reference object from the Pydantic schema name

    characters in the schema name that are invalid/problematic
    for JSONschema $ref names will get replaced with underscores.
    Especially needed for Pydantic generic Models with brackets "[]"

    see: https://github.com/pydantic/pydantic/blob/aee6057378ccfec02126bf9c984a9b6d6b411777/pydantic/json_schema.py#L2031
    """
    ref_name = re.sub(
        r"[^a-zA-Z0-9.\-_]", "_", pydantic_schema.schema_class.__name__
    ).replace(".", "__")
    ref_obj = Reference(**{"$ref": ref_prefix + ref_name})
    logger.debug(f"ref_obj={ref_obj}")
    return ref_obj
